/*
 * light_controller.c
 *
 * CCS C compiler
 * Lighting controller. Reads wireless data from sensor,
 * gets RGB targets, applies PI controller, and sets PWM signals on LEDs.
 *
 *      Authors: Sichao Wang, Jered Greenspan
 *      Copyright 2012
 *
 */

#include <24HJ128GP202.h>
#device *= 16
#fuses HS, NOWDT, NOPROTECT, PUT2
#fuses NODEBUG, NOWRT
#use delay(clock=24.576M,xtal)
#use rs232(baud=1000, rcv=PIN_B6, xmit=PIN_B0,bits=8,parity=N)
// setup PWM outputs
#pin_select OC1=PIN_B7
#pin_select OC2=PIN_B8
#pin_select OC3=PIN_B9

// Matrix to convert XYZ to RGB
#define M11 1.39198
#define M12 -0.208508
#define M13 -0.234335
#define M21 -0.18636
#define M22 0.56342
#define M23 0.133722
#define M31 0.0959742
#define M32 0.107127
#define M33 0.443458

#define SCALING 40
#define ErrorThreshold 10
#define rfactor 1.5
#define gfactor 1
#define bfactor 1.41667

char cct;
char cct_new;
char bright;
char bright_new;
char light_on;
int8 i;

typedef struct rgb_pi {
	char Output;
	int16 Target;
	int16 Sensor;
	int16 Error;
	int8 Saturate;
	int8 Stable;
	int32 ErrorSum;
	float Ki;
	float Kp;
	float newPWM;
} RGB;

RGB pi[3]; // PI controller for each color channel

void readPacket() {
	byte c1, c2;
	char r1;
	char r2;
	char g1;
	char g2;
	char b1;
	char b2;
	c1 = getc();
	c2 = getc();
	while (c1 != 255 || c2 != 255) // get FF header
	{
		c1 = c2;
		c2 = getc();
	}
	r1 = 0;
	r2 = 0;
	g1 = 0;
	g2 = 0;
	b1 = 0;
	b2 = 0;
	r1 = getc();
	r2 = getc();
	g1 = getc();
	g2 = getc();
	b1 = getc();
	b2 = getc();
	cct_new = getc();
	getc();
	bright_new = getc();
	light_on = getc();

	if (cct_new != cct || bright_new != bright) {
		for (i = 0; i < 3; i++)
			pi[i].ErrorSum = 0;
	}
	cct = cct_new;
	bright = bright_new;
	pi[0].Sensor = r1 + r2 * 256;
	pi[1].Sensor = g1 + g2 * 256;
	pi[2].Sensor = b1 + b2 * 256;
}

void getTargetRGB() {
	float x_set;
	float y_set;
	float z_set;
	y_set = ((float) bright) * 30;
	switch (cct) {
	case 2:
		x_set = (1.2749 * y_set);
		z_set = (0.14469 * y_set);
		break;
	case 3:
		x_set = 1.0802 * y_set;
		z_set = 0.39386 * y_set;
		break;
	case 4:
		x_set = 1.0101 * y_set;
		z_set = 0.64454 * y_set;
		break;
	case 5:
		x_set = 0.98123 * y_set;
		z_set = 0.86291 * y_set;
		break;
	case 6:
		x_set = 0.97046 * y_set;
		z_set = 1.0434 * y_set;
		break;
	case 7:
		x_set = 0.96778 * y_set;
		z_set = 1.1908 * y_set;
		break;
	case 8:
		x_set = 0.9685 * y_set;
		z_set = 1.3123 * y_set;
		break;
	case 9:
		x_set = 0.97091 * y_set;
		z_set = 1.412 * y_set;
		break;
	}
	pi[0].Target = (int32)(M11 * x_set + M12 * y_set + M13 * z_set);
	pi[1].Target = (int32)(M21 * x_set + M22 * y_set + M23 * z_set);
	pi[2].Target = (int32)(M31 * x_set + M32 * y_set + M33 * z_set);
}

void CalcPI_RGB(int8 i) {
	float PIout;

	pi[i].Error = pi[i].Target - pi[i].Sensor;
	if (pi[i].Error > 10 || pi[i].Error < -10) {
		pi[i].Stable = 0;
		if (pi[i].Saturate == 0)
			pi[i].ErrorSum += pi[i].Error;
		PIout = pi[i].Kp * pi[i].Error + pi[i].Ki * pi[i].ErrorSum;
		pi[i].Saturate = 0x00;
		pi[i].newPWM = PIout + (float) pi[i].Output;
		/* Debugging:
		printf("[%u]P:%f,I:%f\r\n", i, pi[i].Kp * pi[i].Error,
					pi[i].Ki * pi[i].ErrorSum);
		printf("PIout:%f \r\n",PIout);
		printf("NEWPWM:%f\r\n",pi[i].newPWM);
		*/
		if (pi[i].newPWM <= 0) {
			pi[i].Saturate = 0x01;
			pi[i].Output = 0;
			pi[i].newPWM = 0;
		} else if (pi[i].newPWM >= 255) {
			pi[i].Saturate = 0x02;
			pi[i].Output = 255;
			pi[i].newPWM = 255;
		} else {
			pi[i].Output = (char) (pi[i].newPWM + 0.5);
		}
	} else {
		pi[i].Stable = 1;
		pi[i].ErrorSum = 0;
	}
}

void turnOffLight() {
	set_pwm_duty(1, 8191);
	set_pwm_duty(2, 8191);
	set_pwm_duty(3, 8191);
}

void main() {
	int16 PWM[3]; //buffer to store PWM output

	setup_timer2(TMR_INTERNAL | TMR_DIV_BY_8, 8191);
	setup_compare(1, COMPARE_PWM | COMPARE_TIMER2);
	setup_compare(2, COMPARE_PWM | COMPARE_TIMER2);
	setup_compare(3, COMPARE_PWM | COMPARE_TIMER2);
	setup_adc_ports(NO_ANALOGS);

	for (i = 0; i < 3; i++) {
		pi[i].Output = 0;
		pi[i].Stable = 0;
		pi[i].ErrorSum = 0;
	}

	pi[0].Kp = 0.002;
	pi[1].Kp = 0.012;
	pi[2].Kp = 0.011;
	pi[0].Ki = 0.0001;
	pi[1].Ki = 0.002;
	pi[2].Ki = 0.001;

	delay_ms(1000);

	while (true) {
		if (input(PIN_B5)) { // read PIR sensor
			readPacket();
			if (light_on) {
				getTargetRGB();
				// Apply PI control on each variable
				for (i = 0; i < 3; i++) {
					CalcPI_RGB(i);
				}
				for (i = 0; i < 3; i++) {
					PWM[i] = (255 - pi[i].Output) * (int16) 32;
				}
				/* Debugging:
				printf("Sensor:R:%lu,G:%lu,B:%lu\r\n", pi[0].Sensor,pi[1].Sensor, pi[2].Sensor);
				printf("TARGET:R:%lu,G:%lu,B:%lu\r\n ",pi[0].Target,pi[1].Target,pi[2].Target);
				printf("OUPUT:R:%lu,G:%lu,B:%lu\r\n ",pi[0].Output,pi[1].Output,pi[2].Output);
				*/
				if (pi[0].Stable == 0)
					set_pwm_duty(1, PWM[0]);
				if (pi[1].Stable == 0)
					set_pwm_duty(2, PWM[1]);
				if (pi[2].Stable == 0)
					set_pwm_duty(3, PWM[2]);
			} else {
				turnOffLight();
			}
		} else {
			turnOffLight();
		}

	}
}
